import binascii
from typing import Optional, Union

from django.core import exceptions
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from eth_typing import ChecksumAddress, HexAddress, HexStr
from eth_utils import to_normalized_address
from hexbytes import HexBytes

from ...util.util import to_0x_hex_str
from ..utils import fast_bytes_to_checksum_address, fast_to_checksum_address
from .forms import EthereumAddressFieldForm, HexFieldForm, Keccak256FieldForm
from .validators import validate_address, validate_checksumed_address

try:
    from django.db import DefaultConnectionProxy

    connection = DefaultConnectionProxy()
except ImportError:
    from django.db import connections

    connection = connections["default"]


class EthereumAddressBinaryField(models.Field):
    """
    Stores Ethereum Addresses in binary. Requires keccak256 hashing to
    calculate the EIP55 checksummed address, and it can take a high impact
    on the CPU for a lot of addresses.
    """

    default_validators = [validate_checksumed_address]
    description = "Stores Ethereum Addresses (EIP55) in binary"
    default_error_messages = {
        "invalid": _('"%(value)s" value must be an EIP55 checksummed address.'),
    }

    def get_internal_type(self):
        return "BinaryField"

    def from_db_value(
        self, value: memoryview, expression, connection
    ) -> Optional[ChecksumAddress]:
        if value:
            return fast_bytes_to_checksum_address(value)
        return None

    def get_prep_value(self, value: ChecksumAddress) -> Optional[bytes]:
        if value:
            try:
                return HexBytes(to_normalized_address(value))
            except (TypeError, ValueError):
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )
        return None

    def to_python(self, value) -> Optional[ChecksumAddress]:
        if value is not None:
            try:
                return fast_to_checksum_address(value)
            except ValueError:
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )
        return None

    def formfield(self, **kwargs):
        defaults = {
            "form_class": EthereumAddressFieldForm,
            "max_length": 2 + 40,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class EthereumAddressFastBinaryField(EthereumAddressBinaryField):
    """
    Stores Ethereum Addresses in binary. It returns not EIP55 regular addresses,
    which is faster as not EIP55 checksum is involved.
    """

    default_validators = [validate_address]
    description = "Stores Ethereum Addresses in binary"
    default_error_messages = {
        "invalid": _('"%(value)s" value must be a valid address.'),
    }

    def from_db_value(
        self, value: memoryview, expression, connection
    ) -> Optional[ChecksumAddress]:
        if value:
            return ChecksumAddress(HexAddress(HexStr(to_0x_hex_str(bytes(value)))))

        return None

    def to_python(self, value) -> Optional[ChecksumAddress]:
        if value is not None:
            try:
                if isinstance(value, bytes):
                    if len(value) != 20:
                        raise ValueError(
                            "Cannot convert %s to a checksum address, 20 bytes were expected"
                        )
                return ChecksumAddress(
                    HexAddress(HexStr(to_normalized_address(value)[2:]))
                )
            except ValueError:
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )
        return None


class UnsignedDecimal(models.DecimalField):
    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_digits"]
        del kwargs["decimal_places"]
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return int(value)

    def pre_save(self, model_instance, add):
        """
        Override pre_save to ensure that field is unsigned before save it
        :param model_instance:
        :param add:
        :return:
        """
        value = getattr(model_instance, self.attname)
        if value is not None and value < 0:
            raise ValidationError("Value must be an unsigned 256-bit integer")
        return super().pre_save(model_instance, add)


class Uint256Field(UnsignedDecimal):
    """
    Field to store ethereum uint256 values. Uses Decimal db type without decimals to store
    in the database, but retrieve as `int` instead of `Decimal` (https://docs.python.org/3/library/decimal.html)
    """

    description = _("Ethereum uint256 number")

    def __init__(self, *args, **kwargs):
        kwargs["max_digits"] = 78  # 2 ** 256 is 78 digits
        kwargs["decimal_places"] = 0
        super().__init__(*args, **kwargs)


class Uint96Field(UnsignedDecimal):
    """
    Field to store ethereum uint96 values. Uses Decimal db type without decimals to store
    in the database, but retrieve as `int` instead of `Decimal` (https://docs.python.org/3/library/decimal.html)
    """

    description = _("Ethereum uint96 number")

    def __init__(self, *args, **kwargs):
        kwargs["max_digits"] = 29  # 2 ** 96 is 29 digits
        kwargs["decimal_places"] = 0
        super().__init__(*args, **kwargs)


class Uint32Field(UnsignedDecimal):
    """
    Field to store ethereum uint32 values. Uses Decimal db type without decimals to store
    in the database, but retrieve as `int` instead of `Decimal` (https://docs.python.org/3/library/decimal.html)
    """

    description = _("Ethereum uint32 number")

    def __init__(self, *args, **kwargs):
        kwargs["max_digits"] = 10  # 2 ** 32 is 10 digits
        kwargs["decimal_places"] = 0
        super().__init__(*args, **kwargs)


class HexV2Field(models.BinaryField):
    def formfield(self, **kwargs):
        defaults = {
            "form_class": HexFieldForm,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class Keccak256Field(models.BinaryField):
    description = "Keccak256 hash stored as binary"
    default_error_messages = {
        "invalid": _('"%(value)s" hash must be a 32 bytes hexadecimal.'),
        "length": _('"%(value)s" hash must have exactly 32 bytes.'),
    }

    def _to_bytes(self, value) -> Optional[bytes]:
        if value is None:
            return None
        else:
            try:
                result = HexBytes(value)
                if len(result) != 32:
                    raise exceptions.ValidationError(
                        self.error_messages["length"],
                        code="length",
                        params={"value": value},
                    )
                return result
            except (ValueError, binascii.Error):
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )

    def from_db_value(self, value: memoryview, expression, connection) -> Optional[str]:
        if value:
            return to_0x_hex_str(bytes(value))
        return None

    def get_prep_value(self, value: Union[bytes, str]) -> Optional[bytes]:
        if value:
            return self._to_bytes(value)
        return None

    def value_to_string(self, obj):
        return str(self.value_from_object(obj))

    def to_python(self, value) -> Optional[bytes]:
        if value is not None:
            try:
                return self._to_bytes(value)
            except (ValueError, binascii.Error):
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                )
        return None

    def formfield(self, **kwargs):
        defaults = {
            "form_class": Keccak256FieldForm,
            "max_length": 2 + 64,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


# --------- DEPRECATED, only for old migrations ----------------------
class EthereumAddressField(models.CharField):
    system_check_removed_details = {
        "msg": (
            "EthereumAddressField has been removed except for support in "
            "historical migrations."
        ),
        "hint": "Use EthereumAddressFastBinaryField instead.",
        "id": "fields.E4815",  # pick a unique ID for your field.
    }


class EthereumAddressV2Field(EthereumAddressBinaryField):
    system_check_removed_details = {
        "msg": (
            "EthereumAddressV2Field has been removed except for support in "
            "historical migrations."
        ),
        "hint": "Use EthereumAddressBinaryField instead.",
        "id": "fields.E4816",  # pick a unique ID for your field.
    }


class Sha3HashField(models.CharField):
    system_check_removed_details = {
        "msg": (
            "Sha3HashField has been removed except for support in "
            "historical migrations."
        ),
        "hint": "Use Keccak256Field instead.",
        "id": "fields.E4817",  # pick a unique ID for your field.
    }


class HexField(models.CharField):
    system_check_removed_details = {
        "msg": (
            "HexField has been removed except for support in " "historical migrations."
        ),
        "hint": "Use HexV2Field instead.",
        "id": "fields.E4818",  # pick a unique ID for your field.
    }



import os
import unittest

from django.test import TestCase

import pytest

from ... import EthereumNetwork
from ...clients import EtherscanClientV2, EtherscanRateLimitError
from ...clients.etherscan_client_v2 import AsyncEtherscanClientV2
from .mocks import sourcify_safe_metadata


class TestEtherscanClientV2(TestCase):
    @classmethod
    def get_etherscan_api(cls, network: EthereumNetwork):
        etherscan_api_key_variable_name = "ETHERSCAN_API_KEY"
        etherscan_api_key = os.environ.get(etherscan_api_key_variable_name)
        if not etherscan_api_key:
            pytest.skip(f"{etherscan_api_key_variable_name} needs to be defined")

        return EtherscanClientV2(network, api_key=etherscan_api_key)

    @pytest.mark.flaky(reruns=5)
    def test_etherscan_get_abi(self):
        try:
            etherscan_api = self.get_etherscan_api(EthereumNetwork.MAINNET)
            safe_master_copy_abi = sourcify_safe_metadata["output"]["abi"]
            safe_master_copy_address = "0x6851D6fDFAfD08c0295C392436245E5bc78B0185"
            self.assertEqual(
                etherscan_api.get_contract_abi(safe_master_copy_address),
                safe_master_copy_abi,
            )

            contract_metadata = etherscan_api.get_contract_metadata(
                safe_master_copy_address
            )
            self.assertEqual(contract_metadata.name, "GnosisSafe")
            self.assertEqual(contract_metadata.abi, safe_master_copy_abi)
            self.assertIsNone(contract_metadata.implementation)

            random_address = "0xaE32496491b53841efb51829d6f886387708F99a"
            self.assertIsNone(etherscan_api.get_contract_abi(random_address))
            self.assertIsNone(etherscan_api.get_contract_metadata(random_address))
        except EtherscanRateLimitError:
            self.skipTest("Etherscan rate limit reached")

    @pytest.mark.flaky(reruns=5)
    def test_etherscan_get_contract_metadata(self):
        try:
            etherscan_api = self.get_etherscan_api(EthereumNetwork.MAINNET)
            proxy_address = "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48"

            contract_metadata = etherscan_api.get_contract_metadata(proxy_address)
            self.assertEqual(contract_metadata.name, "FiatTokenProxy")
            self.assertEqual(
                contract_metadata.implementation,
                "0x43506849d7c04f9138d1a2050bbf3a0c054402dd",
            )

        except EtherscanRateLimitError:
            self.skipTest("Etherscan rate limit reached")

    @pytest.mark.flaky(reruns=5)
    def test_is_supported_network(self):
        try:
            self.assertTrue(
                EtherscanClientV2.is_supported_network(EthereumNetwork.GNOSIS)
            )
            self.assertFalse(
                EtherscanClientV2.is_supported_network(EthereumNetwork.UNKNOWN)
            )
        except EtherscanRateLimitError:
            self.skipTest("Etherscan rate limit reached")


class TestAsyncEtherscanClientV2(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def get_etherscan_api(cls, network: EthereumNetwork):
        etherscan_api_key_variable_name = "ETHERSCAN_API_KEY"
        etherscan_api_key = os.environ.get(etherscan_api_key_variable_name)
        if not etherscan_api_key:
            pytest.skip(f"{etherscan_api_key_variable_name} needs to be defined")

        return AsyncEtherscanClientV2(network, api_key=etherscan_api_key)

    @pytest.mark.flaky(reruns=5)
    async def test_async_etherscan_get_abi(self):
        try:
            etherscan_api = self.get_etherscan_api(EthereumNetwork.MAINNET)
            safe_master_copy_abi = sourcify_safe_metadata["output"]["abi"]
            safe_master_copy_address = "0x6851D6fDFAfD08c0295C392436245E5bc78B0185"
            self.assertEqual(
                await etherscan_api.async_get_contract_abi(safe_master_copy_address),
                safe_master_copy_abi,
            )

            contract_metadata = await etherscan_api.async_get_contract_metadata(
                safe_master_copy_address
            )
            self.assertEqual(contract_metadata.name, "GnosisSafe")
            self.assertEqual(contract_metadata.abi, safe_master_copy_abi)

            random_address = "0xaE32496491b53841efb51829d6f886387708F99a"
            self.assertIsNone(
                await etherscan_api.async_get_contract_abi(random_address)
            )
            self.assertIsNone(
                await etherscan_api.async_get_contract_metadata(random_address)
            )
        except EtherscanRateLimitError:
            self.skipTest("Etherscan rate limit reached")



from typing import Any, Iterable, Sequence

from eth_typing import HexStr


def chunks(elements: Sequence[Any], n: int) -> Iterable[Any]:
    """
    :param elements: List
    :param n: Number of elements per chunk
    :return: Yield successive n-sized chunks from l
    """
    for i in range(0, len(elements), n):
        yield elements[i : i + n]


def to_0x_hex_str(value: bytes) -> HexStr:
    """
    Convert bytes to a 0x-prefixed hex string

    :param value: bytes value
    :return: 0x-prefixed hex string
    """
    return HexStr("0x" + value.hex())



import json
import os
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urljoin

import aiohttp
import requests

from safe_eth.eth import EthereumNetwork
from safe_eth.eth.clients import (
    ContractMetadata,
    EtherscanClient,
    EtherscanRateLimitError,
)


class EtherscanClientV2(EtherscanClient):
    """
    Etherscan API V2 supports multiple chains in the same url.

    Reference: https://docs.etherscan.io/etherscan-v2
    """

    BASE_API_V2_URL = "https://api.etherscan.io"

    def __init__(
        self,
        network: EthereumNetwork,
        api_key: Optional[str] = None,
        request_timeout: int = int(
            os.environ.get("ETHERSCAN_CLIENT_REQUEST_TIMEOUT", 10)
        ),
    ):
        super().__init__(EthereumNetwork.MAINNET, api_key, request_timeout)
        self.network = network
        self.base_api_url = self.BASE_API_V2_URL

    def build_url(self, query: str) -> str:
        url = urljoin(self.base_api_url, f"v2/api?chainid={self.network.value}&{query}")
        if self.api_key:
            url += f"&apikey={self.api_key}"
        return url

    @classmethod
    def get_supported_networks(cls) -> List[Dict[str, Any]]:
        """
        Fetches a list of supported networks by the Etherscan API v2.

        :return: List of supported networks, or empty list if request fails.

        Example response
        ```
        {
            "chainname":"Ethereum Mainnet",
            "chainid":"1",
            "blockexplorer":"https://etherscan.io",
            "apiurl":"https://api.etherscan.io/v2/api?chainid=1",
            "status":1
        },
        {
            "chainname":"Sepolia Testnet",
            "chainid":"11155111",
            "blockexplorer":"https://sepolia.etherscan.io",
            "apiurl":"https://api.etherscan.io/v2/api?chainid=11155111",
            "status":1
        }
        ```
        """
        url = urljoin(cls.BASE_API_V2_URL, "v2/chainlist")
        response = requests.get(url)
        if response.ok:
            return response.json().get("result", [])
        return []

    @classmethod
    def is_supported_network(cls, network: EthereumNetwork) -> bool:
        """
        Checks if a given Ethereum network is supported by the Etherscan API v2.

        :param network: The Ethereum network to check.
        :return: `True` if the network is supported; `False` otherwise.
        """
        supported_networks = cls.get_supported_networks()
        return any(
            item.get("chainid") == str(network.value) for item in supported_networks
        )


class AsyncEtherscanClientV2(EtherscanClientV2):
    def __init__(
        self,
        network: EthereumNetwork,
        api_key: Optional[str] = None,
        request_timeout: int = int(
            os.environ.get("ETHERSCAN_CLIENT_REQUEST_TIMEOUT", 10)
        ),
        max_requests: int = int(os.environ.get("ETHERSCAN_CLIENT_MAX_REQUESTS", 100)),
    ):
        super().__init__(network, api_key, request_timeout)
        self.async_session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit_per_host=max_requests)
        )

    async def _async_do_request(
        self, url: str
    ) -> Optional[Union[Dict[str, Any], List[Any], str]]:
        """
        Async version of _do_request
        """
        async with self.async_session.get(
            url, timeout=self.request_timeout
        ) as response:
            if response.ok:
                response_json = await response.json()
                result = response_json["result"]
                if "Max rate limit reached" in result:
                    # Max rate limit reached, please use API Key for higher rate limit
                    raise EtherscanRateLimitError
                if response_json["status"] == "1":
                    return result
            return None

    async def async_get_contract_source_code(
        self,
        contract_address: str,
    ):
        """
        Asynchronous version of get_contract_source_code
        Does not implement retries

        :param contract_address:
        """
        url = self.build_url(
            f"module=contract&action=getsourcecode&address={contract_address}"
        )
        response = await self._async_do_request(url)  # Returns a list
        return self._process_get_contract_source_code_response(response)

    async def async_get_contract_metadata(
        self, contract_address: str
    ) -> Optional[ContractMetadata]:
        contract_source_code = await self.async_get_contract_source_code(
            contract_address
        )
        if contract_source_code:
            return self._process_contract_metadata(contract_source_code)
        return None

    async def async_get_contract_abi(self, contract_address: str):
        url = self.build_url(
            f"module=contract&action=getabi&address={contract_address}"
        )
        result = await self._async_do_request(url)
        if isinstance(result, dict):
            return result
        elif isinstance(result, str):
            try:
                return json.loads(result)
            except json.JSONDecodeError:
                pass
        return None



from django.test import TestCase

from eth_account import Account

from ..serializers import SafeMultisigEstimateTxSerializer


class TestSerializers(TestCase):
    def test_safe_multisig_tx_estimate_serializer(self):
        safe_address = Account.create().address
        eth_address = Account.create().address
        data = {
            "safe": safe_address,
            "to": None,
            "data": None,
            "value": 1,
            "operation": 0,
        }
        serializer = SafeMultisigEstimateTxSerializer(data=data)

        # To and data cannot be empty
        self.assertFalse(serializer.is_valid())

        data = {
            "safe": safe_address,
            "to": eth_address,
            "data": "0x00",
            "value": 1,
            "operation": 2,
        }
        serializer = SafeMultisigEstimateTxSerializer(data=data)
        # Operation cannot be contract creation and to set
        self.assertFalse(serializer.is_valid())

        # Create Operation has been disabled
        data = {
            "safe": safe_address,
            "to": None,
            "data": "0x00",
            "value": 1,
            "operation": 2,
        }
        serializer = SafeMultisigEstimateTxSerializer(data=data)
        # Operation is not contract creation and to is not empty
        self.assertFalse(serializer.is_valid())



from django.db import models

from ..models import (
    EthereumAddressBinaryField,
    EthereumAddressFastBinaryField,
    Keccak256Field,
    Uint32Field,
    Uint96Field,
    Uint256Field,
)


class EthereumAddressBinary(models.Model):
    value = EthereumAddressBinaryField(null=True)


class EthereumAddressFastBinary(models.Model):
    value = EthereumAddressFastBinaryField(null=True)


class Uint256(models.Model):
    value = Uint256Field(null=True)


class Uint96(models.Model):
    value = Uint96Field(null=True)


class Uint32(models.Model):
    value = Uint32Field(null=True)


class Keccak256Hash(models.Model):
    value = Keccak256Field(null=True)



import logging
from typing import Optional, Union, cast

from eth_typing import ChecksumAddress, HexAddress, HexStr

from safe_eth.eth.clients.cowswap import CowSwapAPI, OrderKind
from safe_eth.eth.clients.cowswap.cow_swap_api import AmountResponse, ErrorResponse

from .. import EthereumClient, EthereumNetworkNotSupported
from .exceptions import CannotGetPriceFromOracle
from .oracles import PriceOracle
from .utils import get_decimals

logger = logging.getLogger(__name__)


class CowswapOracle(PriceOracle):
    """
    CowSwap Oracle implementation

    https://docs.cow.fi/off-chain-services/api
    """

    def __init__(
        self,
        ethereum_client: EthereumClient,
    ):
        """
        :param ethereum_client:
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.api = CowSwapAPI(ethereum_client.get_network())

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if CowSwap is available for the EthereumClient provided, `False` otherwise
        """
        try:
            CowSwapAPI(ethereum_client.get_network())
            return True
        except EthereumNetworkNotSupported:
            return False

    def get_price(
        self, token_address_1: str, token_address_2: Optional[str] = None
    ) -> float:
        token_address_2 = token_address_2 or self.api.weth_address
        if token_address_1 == token_address_2:
            return 1.0
        token_address_1_checksum = ChecksumAddress(HexAddress(HexStr(token_address_1)))
        token_address_2_checksum = ChecksumAddress(HexAddress(HexStr(token_address_2)))
        token_1_decimals = get_decimals(token_address_1_checksum, self.ethereum_client)
        try:
            result: Union[
                AmountResponse, ErrorResponse
            ] = self.api.get_estimated_amount(
                token_address_1_checksum,
                token_address_2_checksum,
                OrderKind.SELL,
                10**token_1_decimals,
            )
            if "buyAmount" in result and "sellAmount" in result:
                result = cast(AmountResponse, result)
                # Decimals needs to be adjusted
                token_2_decimals = get_decimals(
                    token_address_2_checksum, self.ethereum_client
                )
                return (
                    float(result["buyAmount"])
                    / result["sellAmount"]
                    * 10 ** (token_1_decimals - token_2_decimals)
                )

            exception = None
        except IOError as exc:
            exception = exc

        message = (
            f"Cannot get price from CowSwap {{}} "
            f"for token-1={token_address_1} to token-2={token_address_2}"
        )
        logger.debug(message)
        raise CannotGetPriceFromOracle(message) from exception



import os
from typing import Any, Dict, Optional
from urllib.parse import urljoin

import aiohttp
import requests
from eth_typing import ChecksumAddress

from .. import EthereumNetwork
from .contract_metadata import ContractMetadata


class BlockscoutClientException(Exception):
    pass


class BlockScoutConfigurationProblem(BlockscoutClientException):
    pass


class BlockscoutClient:
    NETWORK_WITH_URL = {
        EthereumNetwork.ACALA_NETWORK: "https://blockscout.acala.network/api/v2/",
        EthereumNetwork.ALEPH_ZERO_EVM: "https://evm-explorer.alephzero.org/api/v2/",
        EthereumNetwork.ARBITRUM_BLUEBERRY: "https://arb-blueberry.gelatoscout.com/api/v2/",
        EthereumNetwork.ARTHERA_MAINNET: "https://explorer.arthera.net/api/v2/",
        EthereumNetwork.ARTHERA_TESTNET: "https://explorer-test.arthera.net/api/v2/",
        EthereumNetwork.ASTAR_ZKEVM: "https://astar-zkevm.explorer.startale.com/api/v2/",
        EthereumNetwork.AURORIA_TESTNET: "https://auroria.explorer.stratisevm.com/api/v2/",
        EthereumNetwork.BAHAMUT: "https://api.ftnscan.com/api/v2/",
        EthereumNetwork.BOB_SEPOLIA: "https://bob-sepolia.explorer.gobob.xyz/api/v2/",
        EthereumNetwork.BITROCK_TESTNET: "https://testnetscan.bit-rock.io/api/v2/",
        EthereumNetwork.CONNEXT_SEPOLIA: "https://scan.testnet.everclear.org/api/v2/",
        EthereumNetwork.DODOCHAIN_TESTNET: "https://testnet-scan.dodochain.com/api/v2/",
        EthereumNetwork.EVERCLEAR_MAINNET: "https://scan.everclear.org/api/v2/",
        EthereumNetwork.EVMOS: "https://evm.evmos.org/api/v2/",
        EthereumNetwork.ETHERLINK_MAINNET: "https://explorer.etherlink.com/api/v2/",
        EthereumNetwork.EXSAT_MAINNET: "https://scan.exsat.network/api/v2/",
        EthereumNetwork.EXSAT_TESTNET: "https://scan-testnet.exsat.network/api/v2/",
        EthereumNetwork.FLARE_MAINNET: "https://flare-explorer.flare.network/api/v2/",
        EthereumNetwork.FLARE_TESTNET_COSTON2: "https://coston2-explorer.flare.network/api/v2/",
        EthereumNetwork.GAME7: "https://mainnet.game7.io/api/v2/",
        EthereumNetwork.GAME7_TESTNET: "https://testnet.game7.io/api/v2/",
        EthereumNetwork.GNOSIS: "https://gnosis.blockscout.com/api/v2/",
        EthereumNetwork.GNOSIS_CHIADO_TESTNET: "https://gnosis-chiado.blockscout.com/api/v2/",
        EthereumNetwork.GPT_MAINNET: "https://explorer.gptprotocol.io/api/v2/",
        EthereumNetwork.HAQQ_CHAIN_TESTNET: "https://explorer.testedge2.haqq.network/api/v2/",
        EthereumNetwork.HAQQ_NETWORK: "https://explorer.haqq.network/api/v2/",
        EthereumNetwork.HASHKEY_CHAIN: "https://explorer.hsk.xyz/api/v2/",
        EthereumNetwork.HASHKEY_CHAIN_TESTNET: "https://hashkeychain-testnet-explorer.alt.technology/api/v2/",
        EthereumNetwork.IOTA_EVM: "https://iota-evm.blockscout.com/api/v2/",
        EthereumNetwork.INK_SEPOLIA: "https://explorer-sepolia.inkonchain.com/api/v2/",
        EthereumNetwork.JAPAN_OPEN_CHAIN_MAINNET: "https://mainnet.japanopenchain.org/api/v2/",
        EthereumNetwork.JAPAN_OPEN_CHAIN_TESTNET: "https://explorer.testnet.japanopenchain.org/api/v2/",
        EthereumNetwork.KAIA_KAIROS_TESTNET: "https://baobab.scope.klaytn.com/api/v2/",
        EthereumNetwork.KAIA_MAINNET: "https://scope.klaytn.com/api/v2/",
        EthereumNetwork.KROMA: "https://blockscout.kroma.network/api/v2/",
        EthereumNetwork.KROMA_SEPOLIA: "https://blockscout.sepolia.kroma.network/api/v2/",
        EthereumNetwork.LINEA: "https://api-explorer.linea.build/api/v2/",
        EthereumNetwork.LISK: "https://blockscout.lisk.com/api/v2/",
        EthereumNetwork.LISK_SEPOLIA_TESTNET: "https://sepolia-blockscout.lisk.com/api/v2/",
        EthereumNetwork.LORENZO: "https://scan.lorenzo-protocol.xyz/api/v2/",
        EthereumNetwork.MANTLE: "https://explorer.mantle.xyz/api/v2/",
        EthereumNetwork.MANTLE_SEPOLIA_TESTNET: "https://explorer.sepolia.mantle.xyz/api/v2/",
        EthereumNetwork.MANTLE_TESTNET: "https://explorer.testnet.mantle.xyz/api/v2/",
        EthereumNetwork.MANTLE_TESTNET: "https://explorer.testnet.mantle.xyz/api/v2/",
        EthereumNetwork.MANTA_PACIFIC_MAINNET: "https://pacific-explorer.manta.network/api/v2/",
        EthereumNetwork.METER_MAINNET: "https://scan.meter.io/api/v2/",
        EthereumNetwork.METER_TESTNET: "https://scan-warringstakes.meter.io/api/v2/",
        EthereumNetwork.MODE: "https://explorer.mode.network/api/v2/",
        EthereumNetwork.MODE_TESTNET: "https://sepolia.explorer.mode.network/api/v2/",
        EthereumNetwork.NAL_SEPOLIA_TESTNET: "https://testnet-scan.nal.network/api/v2/",
        EthereumNetwork.NEON_EVM_DEVNET: "https://neon-devnet.blockscout.com/api/v2/",
        EthereumNetwork.NEON_EVM_MAINNET: "https://neon.blockscout.com/api/v2/",
        EthereumNetwork.OP_CELESTIA_RASPBERRY: "https://opcelestia-raspberry.gelatoscout.com/api/v2/",
        EthereumNetwork.OP_SEPOLIA_TESTNET: "https://optimism-sepolia.blockscout.com/api/v2/",
        EthereumNetwork.PLUME_DEVNET: "https://test-explorer.plumenetwork.xyz/api/v2/",
        EthereumNetwork.PLUME_MAINNET: "https://phoenix-explorer.plumenetwork.xyz/api/v2/",
        EthereumNetwork.Q_MAINNET: "https://explorer.q.org/api/v2/",
        EthereumNetwork.Q_TESTNET: "https://explorer.qtestnet.org/api/v2/",
        EthereumNetwork.REDSTONE: "https://explorer.redstone.xyz/api/v2/",
        EthereumNetwork.RE_AL: "https://explorer.re.al/api/v2/",
        EthereumNetwork.REYA_NETWORK: "https://explorer.reya.network/api/v2/",
        EthereumNetwork.ROOTSTOCK_MAINNET: "https://rootstock.blockscout.com/api/v2/",
        EthereumNetwork.ROOTSTOCK_TESTNET: "https://rootstock-testnet.blockscout.com/api/v2/",
        EthereumNetwork.SAAKURU_MAINNET: "https://explorer.saakuru.network/api/v2/",
        EthereumNetwork.SNAXCHAIN: "https://explorer.snaxchain.io/api/v2/",
        EthereumNetwork.SONGBIRD_CANARY_NETWORK: "https://songbird-explorer.flare.network/api/v2/",
        EthereumNetwork.SONGBIRD_TESTNET_COSTON: "https://coston-explorer.flare.network/api/v2/",
        EthereumNetwork.STORY_ODYSSEY_TESTNET: "https://odyssey-testnet-explorer.storyscan.xyz/api/v2/",
        EthereumNetwork.SWELLCHAIN: "https://explorer.swellnetwork.io/api/v2/",
        EthereumNetwork.SWELLCHAIN_TESTNET: "https://swell-testnet-explorer.alt.technology/api/v2/",
        EthereumNetwork.TAIKO_HEKLA_L2: "https://blockscoutapi.hekla.taiko.xyz/api/v2/",
        EthereumNetwork.VANA_MOKSHA_TESTNET: "https://api.moksha.vanascan.io/api/v2/",
        EthereumNetwork.ZETACHAIN_TESTNET: "https://zetachain-athens-3.blockscout.com/api/v2/",
        EthereumNetwork.ZORA: "https://explorer.zora.energy/api/v2/",
        EthereumNetwork.ZORA_SEPOLIA_TESTNET: "https://sepolia.explorer.zora.energy/api/v2/",
    }

    def __init__(
        self,
        network: EthereumNetwork,
        request_timeout: int = int(
            os.environ.get("BLOCKSCOUT_CLIENT_REQUEST_TIMEOUT", 10)
        ),
    ):
        self.network = network
        self.api_url = self.NETWORK_WITH_URL.get(network, "")
        self.request_timeout = request_timeout
        if not self.api_url:
            raise BlockScoutConfigurationProblem(
                f"Network {network.name} - {network.value} not supported"
            )
        self.http_session = requests.Session()

    def build_url(self, path: str):
        return urljoin(self.api_url, path)

    def _do_request(self, url: str) -> Optional[Dict[str, Any]]:
        response = self.http_session.get(url, timeout=10)
        if not response.ok:
            return None

        return response.json()

    @staticmethod
    def _process_contract_metadata(
        contract_data: dict[str, Any]
    ) -> Optional[ContractMetadata]:
        """
        Return a ContractMetadata from BlockScout response

        :param contract_data:
        :return:
        """
        if abi := contract_data.get("abi"):
            name = contract_data["name"]
            implementations: list = contract_data.get("implementations", [])
            return ContractMetadata(
                name,
                abi,
                False,
                implementations[0]["address"] if implementations else None,
            )
        return None

    def get_contract_metadata(
        self, address: ChecksumAddress
    ) -> Optional[ContractMetadata]:
        contract_request = self.build_url(f"smart-contracts/{address}")
        contract_data = self._do_request(contract_request)
        if contract_data:
            return self._process_contract_metadata(contract_data)
        return None


class AsyncBlockscoutClient(BlockscoutClient):
    def __init__(
        self,
        network: EthereumNetwork,
        request_timeout: int = int(
            os.environ.get("BLOCKSCOUT_CLIENT_REQUEST_TIMEOUT", 10)
        ),
        max_requests: int = int(os.environ.get("BLOCKSCOUT_CLIENT_MAX_REQUESTS", 100)),
    ):
        super().__init__(network, request_timeout)
        # Limit simultaneous connections to the same host.
        self.async_session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit_per_host=max_requests)
        )

    async def _async_do_request(self, url: str) -> Optional[Dict[str, Any]]:
        """
        Asynchronous version of _do_request
        """
        async with self.async_session.get(
            url, timeout=self.request_timeout
        ) as response:
            if not response.ok:
                return None

            return await response.json()

    async def async_get_contract_metadata(
        self, address: ChecksumAddress
    ) -> Optional[ContractMetadata]:
        contract_request = self.build_url(f"smart-contracts/{address}")
        contract_data = await self._async_do_request(contract_request)
        if contract_data:
            return self._process_contract_metadata(contract_data)
        return None



import os
from functools import cache
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

import aiohttp

from ...util.http import prepare_http_session
from .. import EthereumNetwork
from ..utils import fast_is_checksum_address
from .contract_metadata import ContractMetadata


class SourcifyClientException(Exception):
    pass


class SourcifyClientConfigurationProblem(Exception):
    pass


class SourcifyClient:
    """
    Get contract metadata from Sourcify. Matches can be full or partial:

      - Full: Both the source files and the metadata files were an exact match between the deployed bytecode
        and the published files.
      - Partial: Source code compiles to the same bytecode and thus the contract behaves in the same way,
        but the source code can be different: variables can have misleading names,
        comments can be different and especially the NatSpec comments could have been modified.

    Reference: https://docs.sourcify.dev/docs/api/
    """

    def __init__(
        self,
        network: EthereumNetwork = EthereumNetwork.MAINNET,
        base_url_api: str = "https://sourcify.dev",
        base_url_repo: str = "https://repo.sourcify.dev/",
        request_timeout: int = int(
            os.environ.get("SOURCIFY_CLIENT_REQUEST_TIMEOUT", 10)
        ),
        max_requests: int = int(os.environ.get("SOURCIFY_CLIENT_MAX_REQUESTS", 100)),
    ):
        self.network = network
        self.base_url_api = base_url_api
        self.base_url_repo = base_url_repo
        self.http_session = prepare_http_session(10, max_requests)
        self.request_timeout = request_timeout
        if not self.is_chain_supported(network.value):
            raise SourcifyClientConfigurationProblem(
                f"Network {network.name} - {network.value} not supported"
            )

    def _get_abi_from_metadata(self, metadata: Dict[str, Any]) -> List[Dict[str, Any]]:
        return metadata["output"]["abi"]

    def _get_name_from_metadata(self, metadata: Dict[str, Any]) -> Optional[str]:
        values = list(metadata["settings"].get("compilationTarget", {}).values())
        if values:
            return values[0]
        return None

    def _do_request(self, url: str) -> Optional[Dict[str, Any]]:
        response = self.http_session.get(url, timeout=self.request_timeout)
        if not response.ok:
            return None

        return response.json()

    def is_chain_supported(self, chain_id: int) -> bool:
        chains = self.get_chains()
        if not chains:
            raise IOError("Cannot get chains for SourcifyClient")
        for chain in chains:
            if not isinstance(chain, dict):
                continue
            chain_id_str = chain.get("chainId")
            if chain_id_str is None:
                continue
            try:
                if chain_id == int(chain_id_str):
                    return True
            except ValueError:
                continue
        return False

    @cache
    def get_chains(self) -> Dict[str, Any]:
        url = urljoin(self.base_url_api, "/server/chains")
        result = self._do_request(url)
        return result or {}

    def _process_contract_metadata(
        self, contract_data: dict[str, Any], match_type: str
    ) -> ContractMetadata:
        """
        Return a ContractMetadata from Sourcify response

        :param contract_data:
        :param match_type:
        :return:
        """
        abi = self._get_abi_from_metadata(contract_data)
        name = self._get_name_from_metadata(contract_data)
        return ContractMetadata(name, abi, match_type == "partial_match")

    def get_contract_metadata(
        self, contract_address: str
    ) -> Optional[ContractMetadata]:
        assert fast_is_checksum_address(
            contract_address
        ), "Expecting a checksummed address"

        for match_type in ("full_match", "partial_match"):
            url = urljoin(
                self.base_url_repo,
                f"/contracts/{match_type}/{self.network.value}/{contract_address}/metadata.json",
            )
            contract_data = self._do_request(url)
            if contract_data:
                return self._process_contract_metadata(contract_data, match_type)
        return None


class AsyncSourcifyClient(SourcifyClient):
    def __init__(
        self,
        network: EthereumNetwork = EthereumNetwork.MAINNET,
        base_url_api: str = "https://sourcify.dev",
        base_url_repo: str = "https://repo.sourcify.dev/",
        request_timeout: int = int(
            os.environ.get("SOURCIFY_CLIENT_REQUEST_TIMEOUT", 10)
        ),
        max_requests: int = int(os.environ.get("SOURCIFY_CLIENT_MAX_REQUESTS", 100)),
    ):
        super().__init__(network, base_url_api, base_url_repo, request_timeout)
        # Limit simultaneous connections to the same host.
        self.async_session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit_per_host=max_requests)
        )

    async def _async_do_request(self, url: str) -> Optional[Dict[str, Any]]:
        """
        Asynchronous version of _do_request
        """
        async with self.async_session.get(
            url, timeout=self.request_timeout
        ) as response:
            if not response.ok:
                return None

            return await response.json()

    async def async_get_contract_metadata(
        self, contract_address: str
    ) -> Optional[ContractMetadata]:
        """
        Asynchronous version of get_contract_metadata
        """
        assert fast_is_checksum_address(
            contract_address
        ), "Expecting a checksummed address"

        for match_type in ("full_match", "partial_match"):
            url = urljoin(
                self.base_url_repo,
                f"/contracts/{match_type}/{self.network.value}/{contract_address}/metadata.json",
            )
            contract_data = await self._async_do_request(url)
            if contract_data:
                return self._process_contract_metadata(contract_data, match_type)
        return None



import json
import os
import time
from typing import Any, Dict, List, MutableMapping, Optional, Union
from urllib.parse import urljoin

from ...util.http import prepare_http_session
from .. import EthereumNetwork
from .contract_metadata import ContractMetadata


class EtherscanClientException(Exception):
    pass


class EtherscanClientConfigurationProblem(Exception):
    pass


class EtherscanRateLimitError(EtherscanClientException):
    pass


class EtherscanClient:
    NETWORK_WITH_URL = {
        EthereumNetwork.MAINNET: "https://etherscan.io",
        EthereumNetwork.RINKEBY: "https://rinkeby.etherscan.io",
        EthereumNetwork.ROPSTEN: "https://ropsten.etherscan.io",
        EthereumNetwork.GOERLI: "https://goerli.etherscan.io",
        EthereumNetwork.BNB_SMART_CHAIN_MAINNET: "https://bscscan.com",
        EthereumNetwork.POLYGON: "https://polygonscan.com",
        EthereumNetwork.POLYGON_ZKEVM: "https://zkevm.polygonscan.com",
        EthereumNetwork.PLATON: "https://devnet2openapi.platon.network/rpc",
        EthereumNetwork.OPTIMISM: "https://optimistic.etherscan.io",
        EthereumNetwork.ARBITRUM_ONE: "https://arbiscan.io",
        EthereumNetwork.ARBITRUM_NOVA: "https://nova.arbiscan.io",
        EthereumNetwork.ARBITRUM_GOERLI: "https://goerli.arbiscan.io",
        EthereumNetwork.AVALANCHE_C_CHAIN: "https://snowtrace.io",
        EthereumNetwork.GNOSIS: "https://gnosisscan.io",
        EthereumNetwork.MOONBEAM: "https://moonscan.io",
        EthereumNetwork.MOONRIVER: "https://moonriver.moonscan.io",
        EthereumNetwork.MOONBASE_ALPHA: "https://moonbase.moonscan.io",
        EthereumNetwork.CRONOS_MAINNET: "https://cronoscan.com",
        EthereumNetwork.CRONOS_TESTNET: "https://testnet.cronoscan.com",
        EthereumNetwork.CELO_MAINNET: "https://celoscan.io",
        EthereumNetwork.BASE_GOERLI_TESTNET: "https://goerli.basescan.org",
        EthereumNetwork.NEON_EVM_DEVNET: "https://devnet.neonscan.org",
        EthereumNetwork.NEON_EVM_MAINNET: "https://neonscan.org",
        EthereumNetwork.SEPOLIA: "https://sepolia.etherscan.io",
        EthereumNetwork.ZKSYNC_MAINNET: "https://era.zksync.network/",
        EthereumNetwork.FANTOM_OPERA: "https://ftmscan.com",
        EthereumNetwork.FANTOM_TESTNET: "https://testnet.ftmscan.com/",
        EthereumNetwork.LINEA: "https://lineascan.build",
        EthereumNetwork.LINEA_GOERLI: "https://goerli.lineascan.build",
        EthereumNetwork.MANTLE: "https://mantlescan.xyz/",
        EthereumNetwork.MANTLE_TESTNET: "https://explorer.testnet.mantle.xyz",
        EthereumNetwork.JAPAN_OPEN_CHAIN_MAINNET: "https://mainnet.japanopenchain.org",
        EthereumNetwork.JAPAN_OPEN_CHAIN_TESTNET: "https://explorer.testnet.japanopenchain.org",
        EthereumNetwork.SCROLL_SEPOLIA_TESTNET: "https://sepolia.scrollscan.com",
        EthereumNetwork.SCROLL: "https://scrollscan.com",
        EthereumNetwork.KROMA: "https://blockscout.kroma.network",
        EthereumNetwork.KROMA_SEPOLIA: "https://blockscout.sepolia.kroma.network",
        EthereumNetwork.BLAST_SEPOLIA_TESTNET: "https://sepolia.blastscan.io",
        EthereumNetwork.FRAXTAL: "https://fraxscan.com",
        EthereumNetwork.BASE: "https://api.basescan.org/",
        EthereumNetwork.BLAST: "https://blastscan.io",
        EthereumNetwork.TAIKO_MAINNET: "https://taikoscan.io",
        EthereumNetwork.BASE_SEPOLIA_TESTNET: "https://sepolia.basescan.org",
        EthereumNetwork.HOLESKY: "https://holesky.etherscan.io",
        EthereumNetwork.LINEA_SEPOLIA: "https://sepolia.lineascan.build",
        EthereumNetwork.METIS_ANDROMEDA_MAINNET: "https://explorer.metis.io",
        EthereumNetwork.DOGECHAIN_MAINNET: "https://explorer.dogechain.dog",
        EthereumNetwork.FUSE_SPARKNET: "https://explorer.fusespark.io",
        EthereumNetwork.PLUME_TESTNET: "https://testnet-explorer.plumenetwork.xyz",
        EthereumNetwork.CHILIZ_CHAIN_MAINNET: "https://scan.chiliz.com",
        EthereumNetwork.GNOSIS_CHIADO_TESTNET: "https://blockscout.chiadochain.net",
        EthereumNetwork.AMOY: "https://amoy.polygonscan.com",
        EthereumNetwork.PGN_PUBLIC_GOODS_NETWORK: "https://explorer.publicgoods.network",
        EthereumNetwork.EDGEWARE_EDGEEVM_MAINNET: "https://edgscan.live",
        EthereumNetwork.BEAR_NETWORK_CHAIN_MAINNET: "https://brnkscan.bearnetwork.net",
        EthereumNetwork.THAICHAIN: "https://exp.thaichain.org",
        EthereumNetwork.FUSE_MAINNET: "https://explorer.fuse.io",
        EthereumNetwork.FLUENCE_TESTNET: "https://blockscout.testnet.fluence.dev",
        EthereumNetwork.PULSECHAIN: "https://scan.pulsechain.com",
        EthereumNetwork.IMMUTABLE_ZKEVM: "https://explorer.immutable.com",
        EthereumNetwork.BAHAMUT: "https://www.ftnscan.com",
        EthereumNetwork.ASSET_CHAIN_TESTNET: "https://scan-testnet.assetchain.org",
        EthereumNetwork.LISK: "https://blockscout.lisk.com",
        EthereumNetwork.NEXI_MAINNET: "https://www.nexiscan.com",
        EthereumNetwork.MINATO: "https://explorer-testnet.soneium.org",
        EthereumNetwork.EOS_EVM_NETWORK_TESTNET: "https://explorer.testnet.evm.eosnetwork.com",
        EthereumNetwork.BITTORRENT_CHAIN_MAINNET: "https://bttcscan.com",
        EthereumNetwork.RSS3_VSL_SEPOLIA_TESTNET: "https://scan.testnet.rss3.io",
        EthereumNetwork.BITKUB_CHAIN: "https://www.bkcscan.com",
        EthereumNetwork.METAL_L2: "https://explorer.metall2.com",
        EthereumNetwork.SHIMMEREVM: "https://explorer.evm.shimmer.network",
        EthereumNetwork.HAQQ_CHAIN_TESTNET: "https://explorer.testedge2.haqq.network",
        EthereumNetwork.OASYS_MAINNET: "https://scan.oasys.games",
        EthereumNetwork.MANTA_PACIFIC_MAINNET: "https://pacific-explorer.manta.network",
        EthereumNetwork.FRAXTAL_TESTNET: "https://holesky.fraxscan.com",
        EthereumNetwork.ACALA_NETWORK: "https://blockscout.acala.network",
        EthereumNetwork.ANCIENT8_TESTNET: "https://scanv2-testnet.ancient8.gg",
        EthereumNetwork.FLARE_TESTNET_COSTON2: "https://coston2-explorer.flare.network",
        EthereumNetwork.PUPPYNET: "https://puppyscan.shib.io",
        EthereumNetwork.ROLLUX_MAINNET: "https://explorer.rollux.com",
        EthereumNetwork.ZKSYNC_SEPOLIA_TESTNET: "https://sepolia-era.zksync.network/",
        EthereumNetwork.ATLETA_OLYMPIA: "https://blockscout.atleta.network",
        EthereumNetwork.CELO_ALFAJORES_TESTNET: "https://celo-alfajores.blockscout.com",
        EthereumNetwork.GRAVITY_ALPHA_MAINNET: "https://explorer.gravity.xyz",
        EthereumNetwork.ANCIENT8: "https://scan.ancient8.gg",
        EthereumNetwork.DCHAIN_TESTNET: "https://dchaintestnet-2713017997578000-1.testnet.sagaexplorer.io",
        EthereumNetwork.SYSCOIN_MAINNET: "https://explorer.syscoin.org",
        EthereumNetwork.TENET: "https://tenetscan.io",
        EthereumNetwork.JIBCHAIN_L1: "https://exp-l1.jibchain.net",
        EthereumNetwork.IOTA_EVM_TESTNET: "https://explorer.evm.testnet.iotaledger.net",
        EthereumNetwork.IOTA_EVM: "https://explorer.evm.iota.org",
        EthereumNetwork.BITKUB_CHAIN_TESTNET: "https://testnet.bkcscan.com",
        EthereumNetwork.ROLLUX_TESTNET: "https://rollux.tanenbaum.io",
        EthereumNetwork.MANTA_PACIFIC_TESTNET: "https://pacific-explorer.testnet.manta.network",
        EthereumNetwork.HAQQ_NETWORK: "https://explorer.haqq.network",
        EthereumNetwork.ACALA_MANDALA_TESTNET_TC9: "https://blockscout.mandala.aca-staging.network",
        EthereumNetwork.BERESHEET_BEREEVM_TESTNET: "https://testnet.edgscan.live",
        EthereumNetwork.SEPOLIA_PGN_PUBLIC_GOODS_NETWORK: "https://explorer.sepolia.publicgoods.network",
        EthereumNetwork.LISK_SEPOLIA_TESTNET: "https://sepolia-blockscout.lisk.com",
        EthereumNetwork.ZORA: "https://explorer.zora.energy",
        EthereumNetwork.SATOSHIVM_ALPHA_MAINNET: "https://svmscan.io",
        EthereumNetwork.MANTLE_SEPOLIA_TESTNET: "https://explorer.sepolia.mantle.xyz/",
        EthereumNetwork.SHIMMEREVM_TESTNET: "https://explorer.evm.testnet.shimmer.network",
        EthereumNetwork.BEVM_MAINNET: "https://scan-mainnet.bevm.io",
        EthereumNetwork.CHILIZ_SPICY_TESTNET: "http://spicy-explorer.chiliz.com",
        EthereumNetwork.DCHAIN: "https://dchain-2716446429837000-1.sagaexplorer.io",
        EthereumNetwork.ZORA_SEPOLIA_TESTNET: "https://sepolia.explorer.zora.energy/",
        EthereumNetwork.BLACKFORT_EXCHANGE_NETWORK_TESTNET: "https://testnet-explorer.blackfort.network",
        EthereumNetwork.OP_SEPOLIA_TESTNET: "https://optimism-sepolia.blockscout.com",
        EthereumNetwork.SATOSHIVM_TESTNET: "https://testnet.svmscan.io",
        EthereumNetwork.LUKSO_TESTNET: "https://explorer.execution.testnet.lukso.network",
        EthereumNetwork.CROSSBELL: "https://scan.crossbell.io",
        EthereumNetwork.LUKSO_MAINNET: "https://explorer.execution.mainnet.lukso.network",
        EthereumNetwork.FLUENCE: "https://blockscout.mainnet.fluence.dev",
        EthereumNetwork.SONGBIRD_CANARY_NETWORK: "https://songbird-explorer.flare.network",
        EthereumNetwork.BLACKFORT_EXCHANGE_NETWORK: "https://explorer.blackfort.network",
        EthereumNetwork.RSS3_VSL_MAINNET: "https://scan.rss3.io",
        EthereumNetwork.MANTA_PACIFIC_SEPOLIA_TESTNET: "https://pacific-explorer.sepolia-testnet.manta.network",
        EthereumNetwork.KARURA_NETWORK: "https://blockscout.karura.network",
        EthereumNetwork.APEX_TESTNET: "https://exp-testnet.apexlayer.xyz",
        EthereumNetwork.MODE_TESTNET: "https://sepolia.explorer.mode.network",
        EthereumNetwork.ARBITRUM_SEPOLIA: "https://sepolia.arbiscan.io",
        EthereumNetwork.SONGBIRD_TESTNET_COSTON: "https://coston-explorer.flare.network",
        EthereumNetwork.FLARE_MAINNET: "https://flare-explorer.flare.network",
        EthereumNetwork.FLUENCE_STAGE: "https://blockscout.stage.fluence.dev",
        EthereumNetwork.Q_TESTNET: "https://explorer.qtestnet.org",
        EthereumNetwork.ARTELA_TESTNET: "https://betanet-scan.artela.network",
        EthereumNetwork.EOS_EVM_NETWORK: "https://explorer.evm.eosnetwork.com",
        EthereumNetwork.SHAPE_SEPOLIA_TESTNET: "https://explorer-sepolia.shape.network",
        EthereumNetwork.SHAPE: "https://shapescan.xyz",
        EthereumNetwork.FASTEX_CHAIN_BAHAMUT_OASIS_TESTNET: "https://oasis.ftnscan.com",
        EthereumNetwork.ASSET_CHAIN_MAINNET: "https://scan.assetchain.org",
        EthereumNetwork.PHOENIX_MAINNET: "https://phoenixplorer.com",
        EthereumNetwork.SNAXCHAIN: "https://explorer.snaxchain.io",
        EthereumNetwork.ZKFAIR_MAINNET: "https://scan.zkfair.io",
        EthereumNetwork.SONIC_MAINNET: "https://sonicscan.org",
        EthereumNetwork.BERACHAIN: "https://berascan.com",
    }

    NETWORK_WITH_API_URL = {
        EthereumNetwork.MAINNET: "https://api.etherscan.io",
        EthereumNetwork.RINKEBY: "https://api-rinkeby.etherscan.io",
        EthereumNetwork.ROPSTEN: "https://api-ropsten.etherscan.io",
        EthereumNetwork.GOERLI: "https://api-goerli.etherscan.io",
        EthereumNetwork.BNB_SMART_CHAIN_MAINNET: "https://api.bscscan.com",
        EthereumNetwork.POLYGON: "https://api.polygonscan.com",
        EthereumNetwork.POLYGON_ZKEVM: "https://api-zkevm.polygonscan.com",
        EthereumNetwork.OPTIMISM: "https://api-optimistic.etherscan.io",
        EthereumNetwork.ARBITRUM_ONE: "https://api.arbiscan.io",
        EthereumNetwork.ARBITRUM_NOVA: "https://api-nova.arbiscan.io",
        EthereumNetwork.ARBITRUM_GOERLI: "https://api-goerli.arbiscan.io",
        EthereumNetwork.ARBITRUM_SEPOLIA: "https://api-sepolia.arbiscan.io",
        EthereumNetwork.AVALANCHE_C_CHAIN: "https://api.snowtrace.io",
        EthereumNetwork.GNOSIS: "https://api.gnosisscan.io",
        EthereumNetwork.MOONBEAM: "https://api-moonbeam.moonscan.io",
        EthereumNetwork.MOONRIVER: "https://api-moonriver.moonscan.io",
        EthereumNetwork.MOONBASE_ALPHA: "https://api-moonbase.moonscan.io",
        EthereumNetwork.CRONOS_MAINNET: "https://api.cronoscan.com",
        EthereumNetwork.CRONOS_TESTNET: "https://api-testnet.cronoscan.com",
        EthereumNetwork.CELO_MAINNET: "https://api.celoscan.io",
        EthereumNetwork.BASE_GOERLI_TESTNET: "https://api-goerli.basescan.org",
        EthereumNetwork.NEON_EVM_DEVNET: "https://devnet-api.neonscan.org",
        EthereumNetwork.NEON_EVM_MAINNET: "https://api.neonscan.org",
        EthereumNetwork.SEPOLIA: "https://api-sepolia.etherscan.io",
        EthereumNetwork.ZKSYNC_MAINNET: "https://api-era.zksync.network",
        EthereumNetwork.FANTOM_OPERA: "https://api.ftmscan.com",
        EthereumNetwork.FANTOM_TESTNET: "https://api-testnet.ftmscan.com",
        EthereumNetwork.LINEA: "https://api.lineascan.build",
        EthereumNetwork.LINEA_GOERLI: "https://api-testnet.lineascan.build",
        EthereumNetwork.MANTLE: "https://api.mantlescan.xyz",
        EthereumNetwork.MANTLE_TESTNET: "https://explorer.testnet.mantle.xyz",
        EthereumNetwork.JAPAN_OPEN_CHAIN_MAINNET: "https://mainnet.japanopenchain.org/api",
        EthereumNetwork.JAPAN_OPEN_CHAIN_TESTNET: "https://explorer.testnet.japanopenchain.org/api",
        EthereumNetwork.SCROLL_SEPOLIA_TESTNET: "https://api-sepolia.scrollscan.com",
        EthereumNetwork.SCROLL: "https://api.scrollscan.com",
        EthereumNetwork.KROMA: "https://blockscout.kroma.network",
        EthereumNetwork.KROMA_SEPOLIA: "https://blockscout.sepolia.kroma.network",
        EthereumNetwork.BLAST_SEPOLIA_TESTNET: "https://api-sepolia.blastscan.io",
        EthereumNetwork.FRAXTAL: "https://api.fraxscan.com",
        EthereumNetwork.BASE: "https://api.basescan.org",
        EthereumNetwork.BLAST: "https://api.blastscan.io",
        EthereumNetwork.TAIKO_MAINNET: "https://api.taikoscan.io",
        EthereumNetwork.BASE_SEPOLIA_TESTNET: "https://api-sepolia.basescan.org",
        EthereumNetwork.HOLESKY: "https://api-holesky.etherscan.io",
        EthereumNetwork.LINEA_SEPOLIA: "https://api-sepolia.lineascan.build",
        EthereumNetwork.METIS_ANDROMEDA_MAINNET: "https://api.routescan.io/v2/network/mainnet/evm/1088/etherscan",
        EthereumNetwork.DOGECHAIN_MAINNET: "https://explorer.dogechain.dog",
        EthereumNetwork.FUSE_SPARKNET: "https://explorer.fusespark.io",
        EthereumNetwork.PLUME_TESTNET: "https://testnet-explorer.plumenetwork.xyz",
        EthereumNetwork.CHILIZ_CHAIN_MAINNET: "https://scan.chiliz.com",
        EthereumNetwork.GNOSIS_CHIADO_TESTNET: "https://blockscout.chiadochain.net",
        EthereumNetwork.AMOY: "https://api-amoy.polygonscan.com",
        EthereumNetwork.PGN_PUBLIC_GOODS_NETWORK: "https://explorer.publicgoods.network",
        EthereumNetwork.EDGEWARE_EDGEEVM_MAINNET: "https://edgscan.live",
        EthereumNetwork.BEAR_NETWORK_CHAIN_MAINNET: "https://brnkscan.bearnetwork.net",
        EthereumNetwork.THAICHAIN: "https://exp.thaichain.org",
        EthereumNetwork.FUSE_MAINNET: "https://explorer.fuse.io",
        EthereumNetwork.FLUENCE_TESTNET: "https://blockscout.testnet.fluence.dev",
        EthereumNetwork.PULSECHAIN: "https://api.scan.pulsechain.com",
        EthereumNetwork.IMMUTABLE_ZKEVM: "https://explorer.immutable.com",
        EthereumNetwork.BAHAMUT: "https://www.ftnscan.com",
        EthereumNetwork.ASSET_CHAIN_TESTNET: "https://scan-testnet.assetchain.org",
        EthereumNetwork.LISK: "https://blockscout.lisk.com",
        EthereumNetwork.NEXI_MAINNET: "https://www.nexiscan.com",
        EthereumNetwork.MINATO: "https://explorer-testnet.soneium.org",
        EthereumNetwork.EOS_EVM_NETWORK_TESTNET: "https://explorer.testnet.evm.eosnetwork.com",
        EthereumNetwork.BITTORRENT_CHAIN_MAINNET: "https://api.bttcscan.com",
        EthereumNetwork.RSS3_VSL_SEPOLIA_TESTNET: "https://scan.testnet.rss3.io",
        EthereumNetwork.BITKUB_CHAIN: "https://www.bkcscan.com",
        EthereumNetwork.METAL_L2: "https://explorer.metall2.com",
        EthereumNetwork.SHIMMEREVM: "https://explorer.evm.shimmer.network",
        EthereumNetwork.HAQQ_CHAIN_TESTNET: "https://explorer.testedge2.haqq.network",
        EthereumNetwork.OASYS_MAINNET: "https://scan.oasys.games",
        EthereumNetwork.MANTA_PACIFIC_MAINNET: "https://pacific-explorer.manta.network",
        EthereumNetwork.FRAXTAL_TESTNET: "https://api-holesky.fraxscan.com",
        EthereumNetwork.ACALA_NETWORK: "https://blockscout.acala.network",
        EthereumNetwork.ANCIENT8_TESTNET: "https://scanv2-testnet.ancient8.gg",
        EthereumNetwork.FLARE_TESTNET_COSTON2: "https://coston2-explorer.flare.network",
        EthereumNetwork.PUPPYNET: "https://puppyscan.shib.io",
        EthereumNetwork.ROLLUX_MAINNET: "https://explorer.rollux.com",
        EthereumNetwork.ZKSYNC_SEPOLIA_TESTNET: "https://api-sepolia-era.zksync.network",
        EthereumNetwork.ATLETA_OLYMPIA: "https://blockscout.atleta.network",
        EthereumNetwork.CELO_ALFAJORES_TESTNET: "https://celo-alfajores.blockscout.com",
        EthereumNetwork.GRAVITY_ALPHA_MAINNET: "https://explorer.gravity.xyz",
        EthereumNetwork.ANCIENT8: "https://scan.ancient8.gg",
        EthereumNetwork.DCHAIN_TESTNET: "https://api-dchaintestnet-2713017997578000-1.testnet.sagaexplorer.io",
        EthereumNetwork.SYSCOIN_MAINNET: "https://explorer.syscoin.org",
        EthereumNetwork.TENET: "https://tenetscan.io",
        EthereumNetwork.JIBCHAIN_L1: "https://exp-l1.jibchain.net",
        EthereumNetwork.IOTA_EVM_TESTNET: "https://explorer.evm.testnet.iotaledger.net",
        EthereumNetwork.IOTA_EVM: "https://explorer.evm.iota.org",
        EthereumNetwork.BITKUB_CHAIN_TESTNET: "https://testnet.bkcscan.com",
        EthereumNetwork.ROLLUX_TESTNET: "https://rollux.tanenbaum.io",
        EthereumNetwork.MANTA_PACIFIC_TESTNET: "https://pacific-explorer.testnet.manta.network",
        EthereumNetwork.HAQQ_NETWORK: "https://explorer.haqq.network",
        EthereumNetwork.ACALA_MANDALA_TESTNET_TC9: "https://blockscout.mandala.aca-staging.network",
        EthereumNetwork.BERESHEET_BEREEVM_TESTNET: "https://testnet.edgscan.live",
        EthereumNetwork.SEPOLIA_PGN_PUBLIC_GOODS_NETWORK: "https://explorer.sepolia.publicgoods.network",
        EthereumNetwork.LISK_SEPOLIA_TESTNET: "https://sepolia-blockscout.lisk.com",
        EthereumNetwork.ZORA: "https://explorer.zora.energy",
        EthereumNetwork.SATOSHIVM_ALPHA_MAINNET: "https://svmscan.io",
        EthereumNetwork.MANTLE_SEPOLIA_TESTNET: "https://explorer.sepolia.mantle.xyz",
        EthereumNetwork.SHIMMEREVM_TESTNET: "https://explorer.evm.testnet.shimmer.network",
        EthereumNetwork.BEVM_MAINNET: "https://scan-mainnet-api.bevm.io",
        EthereumNetwork.CHILIZ_SPICY_TESTNET: "http://spicy-explorer.chiliz.com",
        EthereumNetwork.DCHAIN: "https://api-dchain-2716446429837000-1.sagaexplorer.io",
        EthereumNetwork.ZORA_SEPOLIA_TESTNET: "https://sepolia.explorer.zora.energy",
        EthereumNetwork.BLACKFORT_EXCHANGE_NETWORK_TESTNET: "https://testnet-explorer.blackfort.network",
        EthereumNetwork.OP_SEPOLIA_TESTNET: "https://optimism-sepolia.blockscout.com",
        EthereumNetwork.SATOSHIVM_TESTNET: "https://testnet.svmscan.io",
        EthereumNetwork.LUKSO_TESTNET: "https://api.explorer.execution.testnet.lukso.network",
        EthereumNetwork.CROSSBELL: "https://scan.crossbell.io",
        EthereumNetwork.LUKSO_MAINNET: "https://api.explorer.execution.mainnet.lukso.network",
        EthereumNetwork.FLUENCE: "https://blockscout.mainnet.fluence.dev",
        EthereumNetwork.SONGBIRD_CANARY_NETWORK: "https://songbird-explorer.flare.network",
        EthereumNetwork.BLACKFORT_EXCHANGE_NETWORK: "https://explorer.blackfort.network",
        EthereumNetwork.RSS3_VSL_MAINNET: "https://scan.rss3.io",
        EthereumNetwork.MANTA_PACIFIC_SEPOLIA_TESTNET: "https://pacific-explorer.sepolia-testnet.manta.network",
        EthereumNetwork.KARURA_NETWORK: "https://blockscout.karura.network",
        EthereumNetwork.APEX_TESTNET: "https://exp-testnet.apexlayer.xyz",
        EthereumNetwork.MODE_TESTNET: "https://sepolia.explorer.mode.network",
        EthereumNetwork.SONGBIRD_TESTNET_COSTON: "https://coston-explorer.flare.network",
        EthereumNetwork.FLARE_MAINNET: "https://flare-explorer.flare.network",
        EthereumNetwork.FLUENCE_STAGE: "https://blockscout.stage.fluence.dev",
        EthereumNetwork.Q_TESTNET: "https://explorer.qtestnet.org",
        EthereumNetwork.ARTELA_TESTNET: "https://betanet-scan.artela.network",
        EthereumNetwork.EOS_EVM_NETWORK: "https://explorer.evm.eosnetwork.com",
        EthereumNetwork.SHAPE_SEPOLIA_TESTNET: "https://explorer-sepolia.shape.network",
        EthereumNetwork.SHAPE: "https://shapescan.xyz",
        EthereumNetwork.FASTEX_CHAIN_BAHAMUT_OASIS_TESTNET: "https://oasis.ftnscan.com",
        EthereumNetwork.ASSET_CHAIN_MAINNET: "https://scan.assetchain.org",
        EthereumNetwork.PHOENIX_MAINNET: "https://phoenixplorer.com",
        EthereumNetwork.SNAXCHAIN: "https://explorer.snaxchain.io",
        EthereumNetwork.ZKFAIR_MAINNET: "https://scan.zkfair.io",
        EthereumNetwork.SONIC_MAINNET: "https://api.sonicscan.org",
        EthereumNetwork.BERACHAIN: "https://api.berascan.com",
    }
    HTTP_HEADERS: MutableMapping[str, Union[str, bytes]] = {
        "User-Agent": "curl/7.77.0",
    }

    def __init__(
        self,
        network: EthereumNetwork,
        api_key: Optional[str] = None,
        request_timeout: int = int(
            os.environ.get("ETHERSCAN_CLIENT_REQUEST_TIMEOUT", 10)
        ),
    ):
        self.network = network
        self.api_key = api_key
        self.base_url = self.NETWORK_WITH_URL.get(network, "")
        self.base_api_url = self.NETWORK_WITH_API_URL.get(network, "")
        if not self.base_api_url:
            raise EtherscanClientConfigurationProblem(
                f"Network {network.name} - {network.value} not supported"
            )
        self.http_session = prepare_http_session(10, 100)
        self.http_session.headers = self.HTTP_HEADERS
        self.request_timeout = request_timeout

    def build_url(self, query: str):
        url = urljoin(self.base_api_url, f"api?{query}")
        if self.api_key:
            url += f"&apikey={self.api_key}"
        return url

    def _do_request(self, url: str) -> Optional[Union[Dict[str, Any], List[Any], str]]:
        response = self.http_session.get(url, timeout=self.request_timeout)

        if response.ok:
            response_json = response.json()
            result = response_json["result"]
            if "Max rate limit reached" in result:
                # Max rate limit reached, please use API Key for higher rate limit
                raise EtherscanRateLimitError
            if response_json["status"] == "1":
                return result
        return None

    def _retry_request(
        self, url: str, retry: bool = True
    ) -> Optional[Union[Dict[str, Any], List[Any], str]]:
        for _ in range(3):
            try:
                return self._do_request(url)
            except EtherscanRateLimitError as exc:
                if not retry:
                    raise exc
                else:
                    time.sleep(5)
        return None

    @staticmethod
    def _process_contract_metadata(
        contract_data: Dict[str, Any]
    ) -> Optional[ContractMetadata]:
        contract_name = contract_data["ContractName"]
        contract_abi = contract_data["ABI"]
        contract_proxy_implementation_address = (
            contract_data.get("Implementation") or None
        )
        if contract_abi:
            return ContractMetadata(
                contract_name,
                contract_abi,
                False,
                contract_proxy_implementation_address,
            )
        return None

    def get_contract_metadata(
        self, contract_address: str, retry: bool = True
    ) -> Optional[ContractMetadata]:
        contract_source_code = self.get_contract_source_code(
            contract_address, retry=retry
        )
        if contract_source_code:
            return self._process_contract_metadata(contract_source_code)
        return None

    @staticmethod
    def _process_get_contract_source_code_response(response):
        if response and isinstance(response, list):
            result = response[0]
            abi_str = result.get("ABI")

            if isinstance(abi_str, str) and abi_str.startswith("["):
                try:
                    result["ABI"] = json.loads(abi_str)
                except json.JSONDecodeError:
                    result["ABI"] = None  # Handle the case where JSON decoding fails
            else:
                result["ABI"] = None

            return result

    def get_contract_source_code(self, contract_address: str, retry: bool = True):
        """
        Get source code for a contract. Source code query also returns:

            - ContractName: "",
            - CompilerVersion: "",
            - OptimizationUsed: "",
            - Runs: "",
            - ConstructorArguments: ""
            - EVMVersion: "Default",
            - Library: "",
            - LicenseType: "",
            - Proxy: "0",
            - Implementation: "",
            - SwarmSource: ""

        :param contract_address:
        :param retry: if ``True``, try again if there's Rate Limit Error
        :return:
        """
        url = self.build_url(
            f"module=contract&action=getsourcecode&address={contract_address}"
        )
        response = self._retry_request(url, retry=retry)  # Returns a list
        return self._process_get_contract_source_code_response(response)

    def get_contract_abi(self, contract_address: str, retry: bool = True):
        url = self.build_url(
            f"module=contract&action=getabi&address={contract_address}"
        )
        result = self._retry_request(url, retry=retry)
        if isinstance(result, dict):
            return result
        elif isinstance(result, str):
            try:
                return json.loads(result)
            except json.JSONDecodeError:
                pass
        return None
